{-# OPTIONS --type-in-type #-}

-- Burali-Forti's paradox in MLTT-ish without W-types
module Paradox where

-- Σ-types
record Σ (A : Set) (B : A → Set) : Set where
  constructor _,_
  field
    fst : A
    snd : B fst
open Σ
infixr 4 _,_

-- Non-dependent version
_×_ : Set → Set → Set
A × B = Σ A λ _ → B

-- Empty type
data ⊥ : Set where

-- Composition
_∘_ : ∀ {A : Set} {B : A → Set} {C : (a : A) → (B a → Set)}
  → (g : {a : A} → (b : B a) → C a b) → (f : (a : A) → B a) → (a : A) → (C a (f a))
g ∘ f = λ x → g (f x)

-- Identity
id : ∀ {A} → A → A
id x = x

------------------------------------------------------------------------------

-- Binary relation
rel : Set → Set
rel A = A → A → Set

-- Is it transitive?
is-trans : (A : Set) → rel A → Set
is-trans A ordA = ∀ c b a → ordA c b → ordA b a → ordA c a

-- Does the induction principle hold?
has-rec : (A : Set) → rel A → Set
has-rec A ordA = (P : A → Set) → (∀ a → (∀ b → ordA b a → P b) → P a) → ∀ a → P a

-- Is it a "tree" order?
is-tree-order : (A : Set) → rel A → Set
is-tree-order A ordA = is-trans A ordA × has-rec A ordA

-- The type of all "tree"-ordered types.
tree-order : Set
tree-order = Σ Set λ A → Σ (rel A) (is-tree-order A)

-- Convenient projection.
#set : tree-order → Set
#set = fst
#ord : ∀ A → rel (#set A)
#ord A = fst (snd A)
#trans : ∀ A → is-trans (#set A) (#ord A)
#trans A = fst (snd (snd A))
#rec : ∀ A → has-rec (#set A) (#ord A)
#rec A = snd (snd (snd A))

-- The induction principle rejects reflexivity.
rec-irr : ∀ A → ∀ a → #ord A a a → ⊥
rec-irr A = #rec A (λ a → #ord A a a → ⊥) (λ a ih refl → ih a refl refl)

-- [embed-up-to A B b] means that [A] can be embedded into [B]
-- with the strict upper bound [b].
embed-up-to : tree-order → (B : tree-order) → #set B → Set
embed-up-to A B b =
  Σ (#set A → #set B) λ f -- the function
  → (∀ a → #ord B (f a) b) -- the proof that it's a upper bound
  × (∀ {a₁ a₂} → #ord A a₁ a₂ → #ord B (f a₁) (f a₂)) -- the proof that it preserves orders

-- [embed A B] means [A] can be embedded into [B] with some upper bound
embed : tree-order → tree-order → Set
embed A B = Σ (#set B) (embed-up-to A B)

-- [embed] is a transitive relation on [tree-order].
embed-trans : is-trans tree-order embed
embed-trans _ _ _ (b↑ , g , g-bound , g-mono) (_ , f , _ , f-mono)
  = f b↑ , f ∘ g , f-mono ∘ g-bound , f-mono ∘ g-mono

-- [embed] admits an induction principle too!
embed-rec : has-rec tree-order embed 
embed-rec P ind A = ind A lemma
  where
    lemma : ∀ B → embed B A → P B
    lemma B (a↑ , embed-up-to-a↑) = #rec A P′ lemmaA a↑ B embed-up-to-a↑
      where
        P′ : #set A → Set
        P′ a↑ = ∀ B → embed-up-to B A a↑ → P B

        lemmaA : ∀ a↑ → (∀ a↑′ → #ord A a↑′ a↑ → P′ a↑′) → P′ a↑
        lemmaA a↑ ih B (f , f-bound , f-mono) = ind B lemmaCB
          where
            lemmaCB : ∀ C → embed C B → P C
            lemmaCB C (b↑ , g , g-bound , g-mono) =
              ih (f b↑) (f-bound b↑) C (f ∘ g , f-mono ∘ g-bound , f-mono ∘ g-mono)

-- As a result, [embed] is a tree-order!
embed-is-tree-order : is-tree-order tree-order embed
embed-is-tree-order = embed-trans , embed-rec

-- So all the tree-ordered types form a tree-order
tree-order-tree-order : tree-order
tree-order-tree-order = tree-order , embed , embed-is-tree-order

-- Initial segment.
init : ∀ A → #set A → Set
init A a = Σ (#set A) λ b → #ord A b a

-- Induced order.
init-order : ∀ A a → rel (init A a)
init-order A a a+₁ a+₂ = #ord A (fst a+₁) (fst a+₂)

-- Induced order is still a tree-order.
init-is-tree-order : ∀ A a → is-tree-order (init A a) (init-order A a)
init-is-tree-order (A , ordA , transA , recA) a = (λ _ _ _ → transA _ _ _) , λ P indA+ b+ →
  recA (λ b → ∀ b<a → P (b , b<a)) (λ b ih b<a → indA+ (b , b<a) (λ c+ c<b → ih (fst c+) c<b (snd c+))) (fst b+) (snd b+)

-- Therefore the initial segment of a tree-ordered type is a tree-ordered type.
init-tree-order : (A : tree-order) → #set A → tree-order
init-tree-order A a = init A a , init-order A a , init-is-tree-order A a

-- Every initial segment can be embedded back to the original tree-ordered type.
embed-init-self : ∀ A a → embed (init-tree-order A a) A
embed-init-self A a = a , fst , snd , λ < → <

-- And smaller initial segments can be embedded into the larger ones.
embed-init-mono : ∀ A {b a} → #ord A b a → embed (init-tree-order A b) (init-tree-order A a)
embed-init-mono A {b} {a} b<a =
  (b , b<a) ,
  (λ b+ → fst b+ , #trans A _ _ _ (snd b+) b<a) ,
  snd ,
  id

-- Therefore, every tree-ordered type can be embedded into the type of all tree-ordered types,
-- where the upper bound is the tree-ordered type in question.
tree-order-is-max : ∀ A → embed A tree-order-tree-order
tree-order-is-max A = A , init-tree-order A , embed-init-self A , embed-init-mono A

-- It means that we can embed tree-order into tree-order itself.
absurd : ⊥
absurd = rec-irr tree-order-tree-order tree-order-tree-order (tree-order-is-max tree-order-tree-order)

